@@ -13,39 +13,53 @@
 # limitations under the License.
 """Server-side implementation of gRPC Asyncio Python."""
 
-from typing import Text, Optional
 import asyncio
+from concurrent.futures import Executor
+from typing import Any, Optional, Sequence
+
 import grpc
-from grpc import _common
+from grpc import _common, _compression
 from grpc._cython import cygrpc
 
+from . import _base_server
+from ._typing import ChannelArgumentType
 
-class Server:
+
+def _augment_channel_arguments(base_options: ChannelArgumentType,
+                               compression: Optional[grpc.Compression]):
+    compression_option = _compression.create_channel_option(compression)
+    return tuple(base_options) + compression_option
+
+
+class Server(_base_server.Server):
     """Serves RPCs."""
 
-    def __init__(self, thread_pool, generic_handlers, interceptors, options,
-                 maximum_concurrent_rpcs, compression):
+    def __init__(self, thread_pool: Optional[Executor],
+                 generic_handlers: Optional[Sequence[grpc.GenericRpcHandler]],
+                 interceptors: Optional[Sequence[Any]],
+                 options: ChannelArgumentType,
+                 maximum_concurrent_rpcs: Optional[int],
+                 compression: Optional[grpc.Compression]):
         self._loop = asyncio.get_event_loop()
-        self._server = cygrpc.AioServer(self._loop, thread_pool,
-                                        generic_handlers, interceptors, options,
-                                        maximum_concurrent_rpcs, compression)
+        self._server = cygrpc.AioServer(
+            self._loop, thread_pool, generic_handlers, interceptors,
+            _augment_channel_arguments(options, compression),
+            maximum_concurrent_rpcs)
 
     def add_generic_rpc_handlers(
             self,
-            generic_rpc_handlers,
-            # generic_rpc_handlers: Iterable[grpc.GenericRpcHandlers]
-    ) -> None:
+            generic_rpc_handlers: Sequence[grpc.GenericRpcHandler]) -> None:
         """Registers GenericRpcHandlers with this Server.
 
         This method is only safe to call before the server is started.
 
         Args:
-          generic_rpc_handlers: An iterable of GenericRpcHandlers that will be
+          generic_rpc_handlers: A sequence of GenericRpcHandlers that will be
           used to service RPCs.
         """
         self._server.add_generic_rpc_handlers(generic_rpc_handlers)
 
-    def add_insecure_port(self, address: Text) -> int:
+    def add_insecure_port(self, address: str) -> int:
         """Opens an insecure port for accepting RPCs.
 
         This method may only be called before starting the server.
@@ -59,7 +73,7 @@ class Server:
         """
         return self._server.add_insecure_port(_common.encode(address))
 
-    def add_secure_port(self, address: Text,
+    def add_secure_port(self, address: str,
                         server_credentials: grpc.ServerCredentials) -> int:
         """Opens a secure port for accepting RPCs.
 
@@ -141,12 +155,12 @@ class Server:
         self._loop.create_task(self._server.shutdown(None))
 
 
-def server(migration_thread_pool=None,
-           handlers=None,
-           interceptors=None,
-           options=None,
-           maximum_concurrent_rpcs=None,
-           compression=None):
+def server(migration_thread_pool: Optional[Executor] = None,
+           handlers: Optional[Sequence[grpc.GenericRpcHandler]] = None,
+           interceptors: Optional[Sequence[Any]] = None,
+           options: Optional[ChannelArgumentType] = None,
+           maximum_concurrent_rpcs: Optional[int] = None,
+           compression: Optional[grpc.Compression] = None):
     """Creates a Server with which RPCs can be serviced.
 
     Args:
@@ -166,7 +180,8 @@ def server(migration_thread_pool=None,
         indicate no limit.
       compression: An element of grpc.compression, e.g.
         grpc.compression.Gzip. This compression algorithm will be used for the
-        lifetime of the server unless overridden. This is an EXPERIMENTAL option.
+        lifetime of the server unless overridden by set_compression. This is an
+        EXPERIMENTAL option.
 
     Returns:
       A Server object.
